"""MIT License

Copyright (c) 2023-present Tanner B. Corcoran

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

from . import _level_container
from typing import Union
import logging
import typing


class Palette:
    """The color palette used by `ColorFormatter`.
    
    """
    def __init__(self,
                 level_colors: _level_container.LevelContainer[Union[str, typing.Iterable[str]]]
                 ) -> None:
        """Note: all colors are defined as un-escaped ANSI codes with no ending, i.e. `x` in
        `\\033]xm`.


        """
        self._set_colors((logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR,
                          logging.CRITICAL), level_colors.iterall())
    
    def _set_colors(self, levels: typing.Iterable[int],
                    colors: typing.Generator[Union[str, typing.Iterable[str], None], None, None]
                    ) -> None:
        """Set `self.colors` given an iterable of logging levels and the generator returned from
        `LevelContainer.iterall()`.
        
        """
        entry = "\033[{}m"
        def format_many(values) -> str:
            try:
                return "".join(entry.format(v) for v in values)
            except Exception:
                raise ValueError("all colors provided in Palette.__init__ must be "
                                 "either a str or Iterable[str]")
        formatted_colors: dict[int, str] = {}
        for level, color in zip(levels, colors):
            if color is None:
                formatted_colors[level] = ""
                continue
            
            if isinstance(color, str):
                formatted_colors[level] = entry.format(color)
                continue
            
            formatted_colors[level] = format_many(color)

        self.colors = formatted_colors

    def enter(self, levelno: int) -> str:
        """Get the color formatting for a given level.
        
        """
        return self.colors.get(levelno, "")

    def exit(self, levelno: int) -> str:
        """Get the closing formatting for a given level (either `\\033[0m` or an empty str).
        
        """
        return "\033[0m" if self.colors.get(levelno, None) else ""


def parse_palettes(levelno: int, palettes: typing.Mapping[str, Palette]) -> dict[str, str]:
    """Parse all palettes into a dict of their corresponding `_c_enter` and `_c_exit` as keys and
    enter/exit values as values in preparation for formatting with the `Formatter`. Used in
    `plogging.Formatter.formatMessage`.
    
    """
    if not palettes:
        return {}
    itemview = palettes.items()
    c_enter_statements = {f"{n}{chr(2)}":p.enter(levelno) for n, p in itemview}
    c_exit_statements = {f"{n}{chr(3)}":p.exit(levelno) for n, p in itemview}
    return c_enter_statements | c_exit_statements



"""MIT License

Copyright (c) 2023-present Tanner B. Corcoran

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

from typing import Union
from . import types
import typing

_valid_levels = (10, 20, 30, 40, 50)
_name_to_level = {
    "CRITICAL": 50,
    "FATAL": 50,
    "ERROR": 40,
    "WARN": 30,
    "WARNING": 30,
    "INFO": 20,
    "DEBUG": 10,
}


def _verify_level(level: types.LoggingLevel) -> int:
    if isinstance(level, int) and not isinstance(level, bool):
        if level not in _valid_levels:
            raise ValueError(f"Unknown level: {level}")
        return level
    
    if isinstance(level, str):
        if level not in _name_to_level:
            raise ValueError(f"Unknown level: {level}")
        return _name_to_level[level]
    
    raise TypeError(f"Invalid type for level: {type(level).__name__}")


class LevelContainer(typing.Generic[types.T]):
    """A container to hold and obtain values for each of the logging levels available in Python's
    standard `logging` library.
    
    """
    def __init__(self, debug: typing.Optional[types.T] = None,
                 info: typing.Optional[types.T] = None,
                 warning: typing.Optional[types.T] = None,
                 error: typing.Optional[types.T] = None,
                 critical: typing.Optional[types.T] = None, *,
                 default: typing.Optional[types.T] = None) -> None:
        """Values may be provided for each of the logging levels as well as a default fallback for
        any levels whose values are set to or remain `None`.
        
        """
        self._debug = debug
        self._info = info
        self._warning = warning
        self._error = error
        self._critical = critical
        self._default = default
        self._levelmap = {
            10: self.debug,
            20: self.info,
            30: self.warning,
            40: self.error,
            50: self.critical
        }
    
    def _get(self, value: typing.Optional[types.T],
             default: Union[types.VT, types.MISSING]) -> typing.Optional[Union[types.T, types.VT]]:
        """Return value or `_default` or `default` or None.
        
        """
        return value or self._default or (default if default is not types.MISSING else None)

    @typing.overload
    def debug(self) -> typing.Optional[types.T]: ...
    @typing.overload
    def debug(self, default: types.VT, /) -> typing.Optional[Union[types.T, types.VT]]: ...
    def debug(self, default: types.VT = types.MISSING,
              /) -> typing.Optional[Union[types.T, types.VT]]:
        """Get the value for the `DEBUG` logging level, with an optional default if both the value
        and container default are undefined.
        
        """
        return self._get(self._debug, default)

    @typing.overload
    def info(self) -> typing.Optional[types.T]: ...
    @typing.overload
    def info(self, default: types.VT, /) -> typing.Optional[Union[types.T, types.VT]]: ...
    def info(self, default: types.VT = types.MISSING,
             /) -> typing.Optional[Union[types.T, types.VT]]:
        """Get the value for the `INFO` logging level, with an optional default if both the value
        and container default are undefined.
        
        """
        return self._get(self._info, default)

    @typing.overload
    def warning(self) -> typing.Optional[types.T]: ...
    @typing.overload
    def warning(self, default: types.VT, /) -> typing.Optional[Union[types.T, types.VT]]: ...
    def warning(self, default: types.VT = types.MISSING,
                /) -> typing.Optional[Union[types.T, types.VT]]:
        """Get the value for the `WARNING` logging level, with an optional default if both the value
        and container default are undefined.
        
        """
        return self._get(self._warning, default)

    @typing.overload
    def error(self) -> typing.Optional[types.T]: ...
    @typing.overload
    def error(self, default: types.VT, /) -> typing.Optional[Union[types.T, types.VT]]: ...
    def error(self, default: types.VT = types.MISSING,
              /) -> typing.Optional[Union[types.T, types.VT]]:
        """Get the value for the `ERROR` logging level, with an optional default if both the value
        and container default are undefined.
        
        """
        return self._get(self._error, default)

    @typing.overload
    def critical(self) -> typing.Optional[types.T]: ...
    @typing.overload
    def critical(self, default: types.VT, /) -> typing.Optional[Union[types.T, types.VT]]: ...
    def critical(self, default: types.VT = types.MISSING,
                 /) -> typing.Optional[Union[types.T, types.VT]]:
        """Get the value for the `CRITICAL` logging level, with an optional default if both the
        value and container default are undefined.
        
        """
        return self._get(self._critical, default)
    
    @typing.overload
    def get(self, level: types.LoggingLevel, /) -> typing.Optional[types.T]: ...
    @typing.overload
    def get(self, level: types.LoggingLevel,
            default: types.VT, /) -> typing.Optional[Union[types.T, types.VT]]: ...
    def get(self, level: types.LoggingLevel,
            default: types.VT = types.MISSING, /) -> typing.Optional[Union[types.T, types.VT]]:
        """Get the value for the the given logging level, with an optional default if both the
        value and container default are undefined.
        
        """
        return self._levelmap[_verify_level(level)](default)

    @typing.overload
    def iterall(self) -> typing.Generator[typing.Optional[types.T], None, None]: ...
    @typing.overload
    def iterall(self, debug_default: types.VT = types.MISSING,
                info_default: types.VT = types.MISSING, warning_default: types.VT = types.MISSING,
                error_default: types.VT = types.MISSING, critical_default: types.VT = types.MISSING
                ) -> typing.Generator[typing.Optional[Union[types.T, types.VT]], None, None]: ...
    def iterall(self, debug_default: types.VT = types.MISSING,
                info_default: types.VT = types.MISSING, warning_default: types.VT = types.MISSING,
                error_default: types.VT = types.MISSING, critical_default: types.VT = types.MISSING
                ) -> typing.Generator[typing.Optional[Union[types.T, types.VT]], None, None]:
        """Iterate through all logging levels, with optional defaults for each level.
        
        """
        funcs = (self.debug, self.info, self.warning, self.error, self.critical)
        defaults = (debug_default, info_default, warning_default, error_default, critical_default)
        for f, d in zip(funcs, defaults):
            yield f(d)



"""MIT License

Copyright (c) 2023-present Tanner B. Corcoran

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

from typing import Union
from . import types
import string
str_formatter = string.Formatter()
del string


class FormatContainer:
    def __init__(self, conversion: str, format_spec: str, field_name: str = None,
                 field_name_append: str = None) -> None:
        self.conversion = conversion
        self.format_spec = format_spec
        self.field_name = field_name
        self.field_name_append = field_name_append
    
    def __str__(self) -> str:
        return f"{{{self.field_name}{self.field_name_append}{self.conversion}{self.format_spec}}}"


class FormatStringConverter:
    def __init__(self) -> None:
        self.values: list[Union[str, FormatContainer]] = []
        self.awaiting_exit: FormatContainer = None

    @staticmethod
    def _conversion(conversion: Union[str, None]) -> str:
        return f"!{conversion}" if conversion else ""

    @staticmethod
    def _format_spec(format_spec: Union[str, None]) -> str:
        return f":{format_spec}" if format_spec else ""

    @staticmethod
    def _format_statement(field_name: Union[str, None], conversion: str, format_spec: str) -> str:
        return f"{{{field_name or ''}{conversion}{format_spec}}}"

    def _add_flag(self, field_name: Union[str, None], conversion: str, format_spec: str) -> None:
        self.values.append(FormatStringConverter._format_statement(field_name, conversion,
                                                                   format_spec))

    def _enter(self, conversion: str, format_spec: str) -> None:
        if self.awaiting_exit:
            raise ValueError("new enter flag found before previous enter flag was closed")
        enter = FormatContainer(conversion, format_spec, field_name_append=chr(2))
        self.values.append(enter)
        self.awaiting_exit = enter
    
    def _exit(self, conversion: str, format_spec: str) -> None:
        if not self.awaiting_exit:
            raise ValueError("exit flag used before enter flag")
        if self.awaiting_exit.field_name is None:
            raise ValueError("exit flag used without a contained flag; make sure you have at least"
                             " one flag between your enter and exit flags")
        self._add_flag(f"{self.awaiting_exit.field_name}{chr(3)}", conversion, format_spec)
        self.awaiting_exit = None
    
    def _stringify_values(self) -> list[str]:
        return [str(v) for v in self.values]

    def to_color_format(self, fmt: types.SupportsBracketFormat) -> str:
        for literal_text, field_name, format_spec, conversion in str_formatter.parse(fmt):
            if literal_text:
                self.values.append(literal_text)
            
            _conversion = self._conversion(conversion)
            _format_spec = self._format_spec(format_spec)

            if field_name is None and format_spec is None and conversion is None:
                continue

            if not field_name:
                self._add_flag(None, _conversion, _format_spec)
                continue
            
            if field_name.startswith("_"):
                self._add_flag(field_name[1:], _conversion, _format_spec)
                continue

            if field_name == "enter":
                self._enter(_conversion, _format_spec)
                continue
            
            if field_name == "exit":
                self._exit(_conversion, _format_spec)
                continue
            
            if self.awaiting_exit and self.awaiting_exit.field_name is not None:
                self._add_flag(field_name, _conversion, _format_spec)
                continue
            
            if self.awaiting_exit:
                self.awaiting_exit.field_name = field_name
                self._add_flag(field_name, _conversion, _format_spec)
                continue
            
            self._add_flag(f"{field_name}{chr(2)}", "", "")
            self._add_flag(field_name, _conversion, _format_spec)
            self._add_flag(f"{field_name}{chr(3)}", "", "")

        values = self._stringify_values()

        self.values = []
        self.awaiting_exit = None
        
        return "".join(values)
    
    @staticmethod
    def strip_color(fmt: types.SupportsBracketFormat) -> str:
        updated: list[str] = []

        for literal_text, field_name, format_spec, conversion in str_formatter.parse(fmt):
            if literal_text:
                updated.append(literal_text)
            
            _conversion = FormatStringConverter._conversion(conversion)
            _format_spec = FormatStringConverter._format_spec(format_spec)

            if field_name in ["enter", "exit"]:
                continue
            
            if field_name and field_name[-1] in [chr(2), chr(3)]:
                continue
            
            if field_name and field_name.startswith("_"):
                field_name = field_name[1:]

            updated.append(FormatStringConverter._format_statement(field_name, _conversion,
                                                                   _format_spec))
        
        return "".join(updated)



"""The core module which defines `Formatter`.

:copyright: (c) 2023-present Tanner B. Corcoran
:license: MIT, see LICENSE for more details.

"""

__author__ = "Tanner B. Corcoran"
__license__ = "MIT License"
__copyright__ = "Copyright (c) 2023-present Tanner B. Corcoran"

from . import _level_container
from . import _converter
from . import constants
from . import _palette
from . import types
import traceback
import logging
import string
import typing
import io
import re

fmt_spec = re.compile(r"^(.?[<>=^])?[+ -]?#?0?(\d+|{\w+})?[,_]?(\.(\d+|{\w+}))?[bcdefgnosx%]?$",
                      re.IGNORECASE)
field_spec = re.compile(r"^(\d+|\w+)(\.\w+|\[[^]]+\])*$")
str_formatter = string.Formatter()


def _validate_fmt(fmt: str, extra_keys: list[str]) -> None:
    """Validate the input format, ensure it is the correct string formatting style.
    
    """
    fields: set[str] = set()
    try:
        for literal_text, field_name, format_spec, conversion in str_formatter.parse(fmt):
            if field_name:
                if not field_spec.match(field_name):
                    raise ValueError(f"invalid field name/expression: {field_name}")
                fields.add(field_name)
            if conversion and conversion not in 'rsa':
                raise ValueError(f"invalid conversion: {conversion}")
            if format_spec and not fmt_spec.match(format_spec):
                raise ValueError(f"bad specifier: {format_spec}")
    except ValueError as exc:
        raise ValueError(f"Invalid format: {exc}")
    if not fields:
        raise ValueError("Invalid format: no fields")
    fields = {f[1:] if f.startswith("_") else f for f in fields}
    missing = fields - set(constants.VALID_FLAGS + list(extra_keys))
    if missing:
        raise ValueError(f"Missing defaults: {', '.join(missing)}")


def _format_defaults(default_keys: list[str]) -> dict:
    if not default_keys:
        return {}
    c_enter_statements = {f"{name}{chr(2)}":"" for name in default_keys}
    c_exit_statements = {f"{name}{chr(3)}":"" for name in default_keys}
    return c_enter_statements | c_exit_statements


class Formatter(logging.Formatter):
    """A `logging.Formatter` subclass equipped to handle both standard and color formatting.
    
    """
    def __init__(self, fmts: _level_container.LevelContainer[types.SupportsBracketFormat],
                 datefmt: str = None, defaults: typing.Mapping = None,
                 force_ansi: bool = False,
                 backup_fmts: _level_container.LevelContainer[types.SupportsBracketFormat] = None,
                 **palettes: _palette.Palette) -> None:
        """
        Arguments
        ---------
        fmts : LevelContainer[SupportsBracketFormat]
            The container specifying format strings for each (or all) of the logging levels. Unlike
            `logging.Formatter`, which can use bracket, percent, and shell formatting, this
            formatter requires the user to use bracket formatting. The list of allowed flags can be
            acquired through the `VALID_FLAGS` module variable.
        datefmt : str, default=None
            The date format string used when formatting the `asctime` flag. This uses the same
            formatting style as `datetime`'s `strftime` and `strptime` methods.
        defaults : Mapping, default=None
            Default values for any flag.
        force_ansi : bool, default=False
            If `True`, the ANSI color coding will be included in the output regardless of whether or
            not the terminal in use supports color.
        backup_fmts : LevelContainer[SupportsBracketFormat], default=None
            The formats used when the current terminal use does not support color. If `None`, the
            main formats will be used (though without colors applied).
        **palettes : Palette
            Additional keyword-only arguments used to set the colors of various flags for different
            logging levels. All keys must be valid flags; a list of which can be acquired through
            the `VALID_FLAGS` module variable.
        
        Notes
        -----
        The `stream_supports_color` attribute must be set to a boolean value post-instantiation.
        This variable indicates whether or not the current stream supports ANSI color formatting.
        When using `setup_existing` or `setup_new`, this is set automatically.

        """
        self.defaults = defaults or {}
        self.default_keys = list(self.defaults.keys())
        self._ensure_sufficient_formats(fmts)
        self._ensure_format_validity(fmts, self.default_keys)
        if backup_fmts:
            self._ensure_format_validity(backup_fmts, self.default_keys)
            self._fix_backups(fmts, backup_fmts, False)
        else:
            backup_fmts = _level_container.LevelContainer()
            self._fix_backups(fmts, backup_fmts, True)
        self._update_formats(fmts)
        

        self.force_ansi = force_ansi
        self.stream_supports_color = None
        self.palettes = palettes
        self.fmts = fmts
        self.backup_fmts = backup_fmts
        
        super().__init__()
    
    @staticmethod
    def _ensure_format_validity(fmt: _level_container.LevelContainer[types.SupportsBracketFormat],
                                extras: "list[str]") -> None:
        """Ensure all formats in `fmt` are valid bracket-style format strings and that all fields
        are in either VALID_FLAGS or `extras`.
        
        """
        for fmt_str in fmt.iterall():
            if fmt_str is not None:
                _validate_fmt(fmt_str, extras)

    @staticmethod
    def _ensure_sufficient_formats(fmt: _level_container.LevelContainer[types.SupportsBracketFormat]
                                   ) -> None:
        """Ensure that all logging levels are covered in `fmt`.
        
        """
        # this case would technically be covered in `.iterall()`, but this is faster
        if fmt._default is not None:
            return

        if not all(x is not None for x in fmt.iterall()):
            raise ValueError("all levels in fmt must have a format string (either set the default "
                             "or provide a value for each level)")

    @staticmethod
    def _update_formats(fmt: _level_container.LevelContainer[types.SupportsBracketFormat]) -> None:
        """Update each format string in `fmt` to a color-ready format string.
        
        """
        converter = _converter.FormatStringConverter()
        for attr in ("_debug", "_info", "_warning", "_error", "_critical", "_default"):
            _fmt: str = getattr(fmt, attr)
            if _fmt is not None:
                setattr(fmt, attr, converter.to_color_format(_fmt))
    
    @staticmethod
    def _fix_backups(fmts: _level_container.LevelContainer[types.SupportsBracketFormat],
                     backup_fmts: _level_container.LevelContainer[types.SupportsBracketFormat],
                     make_new: bool) -> None:
        if make_new:
            for attr in ("_debug", "_info", "_warning", "_error", "_critical", "_default"):
                val = getattr(fmts, attr)
                if val:
                    setattr(backup_fmts, attr, _converter.FormatStringConverter.strip_color(val))
            return
        
        for attr in ("debug", "info", "warning", "error", "critical"):
            val = getattr(backup_fmts, attr)()
            if not val:
                setattr(backup_fmts, f"_{attr}", getattr(fmts, attr)()) 

    def _ensure_ssc_set(self) -> None:
        """Ensure that `stream_supports_color` is set to a boolean value.
        
        """
        if self.stream_supports_color not in [True, False, 1, 0]:
            raise ValueError("ColorFormatter.stream_supports_color must be a boolean")

    def _get_fmt_str(self, levelno: int) -> str:
        """Get the format string from `fmts` or `backup_fmts` given the leveno.
        
        """
        self._ensure_ssc_set()
        
        if self.force_ansi is True or self.stream_supports_color is True:
            return self.fmts.get(levelno)
        
        if self.stream_supports_color is False and self.backup_fmts is None:
            # no colors applied
            return self.fmts.get(levelno)
        
        # if we have a backup format and the stream does not support color
        return self.backup_fmts.get(levelno) or self.fmts.get(levelno)
    
    def _get_fmt_kwargs(self, record: logging.LogRecord) -> dict:
        """Get the kwargs from `fmts` or `backup_fmts` for formatting the format string given the
        `LogRecord`.
        
        """
        self._ensure_ssc_set()

        if self.force_ansi is True or self.stream_supports_color is True:
            return (constants.COLOR_FORMAT_DEFAULTS | _format_defaults(self.default_keys)
                    | _palette.parse_palettes(record.levelno, self.palettes)
                    | self.defaults | record.__dict__)
        
        if self.stream_supports_color is False and self.backup_fmts is None:
            # no colors applied
            return (constants.COLOR_FORMAT_DEFAULTS | _format_defaults(self.default_keys)
                    | self.defaults | record.__dict__)
        
        # if we have a backup format and the stream does not support color
        return self.defaults | record.__dict__

    def formatException(self, exc_info: tuple, leveno: int) -> str:
        with io.StringIO() as sio:
            traceback.print_exception(*exc_info, limit=None, file=sio)
            s = sio.getvalue()
        if s[-1:] == "\n":
            s = s[:-1]
        
        if self.force_ansi is False and self.stream_supports_color is False:
            return s
        
        exc_text_palette = self.palettes.get("exc_text", None)
        if exc_text_palette is None:
            return s
        
        exc_text_c_enter = exc_text_palette.enter(leveno)
        exc_text_c_exit = exc_text_palette.exit(leveno)
        
        return f"{exc_text_c_enter}{s}{exc_text_c_exit}"

    def usesTime(self, fmt: str) -> bool:
        return fmt.find(logging.StrFormatStyle.asctime_search) >= 0

    def formatMessage(self, fmt: str, record: logging.LogRecord) -> str:
        kwargs = self._get_fmt_kwargs(record)
        try:
            return fmt.format(**kwargs)
        except KeyError as exc:
            raise ValueError(f"Formatting field not found in record: {exc!a}")
        
    def format(self, record: logging.LogRecord) -> str:
        record.message = record.getMessage()
        fmt = self._get_fmt_str(record.levelno)
        
        # set record's time only if "{asctime" is found in the format string
        if self.usesTime(fmt):
            record.asctime = self.formatTime(record, self.datefmt)
        
        # initial message format
        s = self.formatMessage(fmt, record)

        # custom format for exceptions (requires record.leveno to be passed)
        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info, record.levelno)

                # remove the cache layer
                record.exc_info = None

        # the rest of this is default
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + self.formatStack(record.stack_info)
        return s



"""MIT License

Copyright (c) 2023-present Tanner B. Corcoran

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

from . import _formatter
from . import _palette
from . import _level_container
from . import types
import logging
import sys
import os
import io


def _stream_supports_color(stream: io.TextIOBase) -> bool:
    """Return whether or not the running system's terminal supports ANSI color formatting.
    
    """

    def vt_codes_enabled_in_windows_registry() -> bool:
        try:
            import winreg
        except ImportError:
            return False
        else:
            try:
                reg_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, "Console")
                reg_key_value, _ = winreg.QueryValueEx(reg_key, "VirtualTerminalLevel")
            except FileNotFoundError:
                return False
            else:
                return reg_key_value == 1

    is_a_tty = hasattr(stream, "isatty") and stream.isatty()

    if sys.platform != "win32":
        return is_a_tty

    return is_a_tty and (
        "ANSICON" in os.environ
        or "PYCHARM_HOSTED" in os.environ # pycharm
        or "WT_SESSION" in os.environ # Windows terminal
        or os.environ.get("TERM_PROGRAM") == "vscode" # vscode
        or vt_codes_enabled_in_windows_registry()
    )


def default_formatter() -> _formatter.Formatter:
    """Create the default `Formatter` instance.
    
    """
    return _formatter.Formatter(
        _level_container.LevelContainer(
            default="{asctime} {levelname:<8} {name} {message}"
        ),
        "%Y-%m-%d %H:%M:%S",
        backup_fmts=_level_container.LevelContainer(
            default="[{asctime}] [{levelname:<8}] {name}: {message}"
        ),
        asctime=_palette.Palette(
            _level_container.LevelContainer(
                default="30;1"
            )
        ),
        levelname=_palette.Palette(
            _level_container.LevelContainer(
                "32;1",
                "34;1",
                "33;1",
                "31;1",
                "41"
            )
        ),
        name=_palette.Palette(
            _level_container.LevelContainer(
                default="35"
            )
        ),
        exc_text=_palette.Palette(
            _level_container.LevelContainer(
                default="31"
            )
        )
    )


def setup_existing(logger: logging.Logger, *, level: types.LoggingLevel = None,
                   handler: logging.Handler = None, formatter: logging.Formatter = None) -> None:
    """Setup formatting and handling on an existing `Logger`.

    Arguments
    ---------
    level : LoggingLevel
        If defined, the logging level for `logger` will be set to `level`.
    handler : Handler
        If defined, the handler for `logger` will be set to `handler`. If not defined, a new
        `StreamHandler` will be used.
    formatter : Formatter
        If defined, the formatter for `handler` will be set to `formatter`. If not defined, a
        default instance of `plogging.Formatter` will be used instead.
    
    Notes
    -----
    If `formatter` is an instance of `plogging.Formatter`, the `stream_supports_color` attribute on
    the formatter will be set to `True` if the handler stream supports color, otherwise `False`. If
    this value is already set, then it will not be changed.
    
    """
    handler = handler or logging.StreamHandler()

    if formatter is None:
        formatter = default_formatter()
    
    if isinstance(formatter, _formatter.Formatter) and formatter.stream_supports_color is None:
        if not hasattr(handler, "stream"):
            stream_supports_color = False
        else:
            stream_supports_color = _stream_supports_color(getattr(handler, "stream"))
        formatter.stream_supports_color = stream_supports_color

    if level:
        logger.setLevel(level)

    handler.setFormatter(formatter)
    logger.addHandler(handler)


def setup_new(name: str, *, level: types.LoggingLevel = None, package: str = None,
              handler: logging.Handler = None,
              formatter: logging.Formatter = None) -> logging.Logger:
    """Setup a new `Logger` with the given `name` and optional `package` (will be concatenated into
    `{package}.{name}). `level`, `handler`, and `formatter` are the same as those found in
    `setup_existing`.
    
    """
    if package:
        _name = f"{package}.{name}"
    else:
        _name = name
    logger = logging.getLogger(_name)
    setup_existing(logger, level=level, handler=handler, formatter=formatter)
    return logger



"""A formatter and set of utilities that bring pretty, highly-customizable and colored log messages
to the standard logging library.

:copyright: (c) 2023-present Tanner B. Corcoran
:license: MIT, see LICENSE for more details.

"""

__title__ = "plogging"
__author__ = "Tanner B. Corcoran"
__email__ = "tannerbcorcoran@gmail.com"
__license__ = "MIT License"
__copyright__ = "Copyright (c) 2023-present Tanner B. Corcoran"
__version__ = "0.0.1"
__description__ = ("A formatter and set of utilities that bring pretty, highly-customizable and "
                   "colored log messages to the standard logging library")
__url__ = "https://github.com/tanrbobanr/plogging"
__download_url__ = "https://pypi.org/project/plogging"


__all__ = (
    "Formatter",
    "LevelContainer",
    "Palette",
    "VALID_FLAGS",
    "COLOR_FORMAT_DEFAULTS",
    "setup_existing",
    "setup_new"
)


from ._formatter import Formatter
from ._level_container import LevelContainer
from ._palette import Palette
from .constants import (
    VALID_FLAGS,
    COLOR_FORMAT_DEFAULTS
)
from .utils import (
    setup_existing,
    setup_new
)
